package org.opengis.cite.iso19142.simple;

import com.sun.jersey.api.client.ClientResponse;
import java.net.URI;
import java.util.Collections;
import java.util.HashMap;
import java.util.Map;
import java.util.UUID;
import javax.xml.namespace.QName;
import javax.xml.transform.dom.DOMResult;
import javax.xml.transform.dom.DOMSource;
import javax.xml.validation.Schema;
import javax.xml.xpath.XPathExpressionException;
import org.opengis.cite.iso19142.ErrorMessage;
import org.opengis.cite.iso19142.ErrorMessageKeys;
import org.opengis.cite.iso19142.BaseFixture;
import org.opengis.cite.iso19142.Namespaces;
import org.opengis.cite.iso19142.ProtocolBinding;
import org.opengis.cite.iso19142.SuiteAttribute;
import org.opengis.cite.iso19142.WFS2;
import org.opengis.cite.iso19142.ETSAssert;
import org.opengis.cite.iso19142.util.XMLUtils;
import org.opengis.cite.iso19142.util.ServiceMetadataUtils;
import org.opengis.cite.iso19142.util.ValidationUtils;
import org.opengis.cite.iso19142.util.WFSRequest;
import org.opengis.cite.validation.SchematronValidator;
import org.testng.Assert;
import org.testng.ITestContext;
import org.testng.annotations.BeforeClass;
import org.testng.annotations.BeforeMethod;
import org.testng.annotations.Test;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;

/**
 * Tests the service response to a GetFeature request that invokes a stored
 * query. A WFS implementation is required to support stored queries at this
 * conformance level.
 * 
 * @see "ISO 19142:2010, cl. A.1.1: Simple WFS"
 * @see "ISO 19142:2010, cl. 11: GetFeature operation"
 * @see "ISO 19142:2010, cl. 7.9.3: Stored query expression"
 */
@Test(groups = "simple")
public class StoredQueryTests extends BaseFixture {

    private Schema wfsSchema;
    /** Map entries are {id: type-name} */
    private Map<String, QName> featureIdentifiers = new HashMap<String, QName>();

    /**
     * Retrieves the (pre-compiled) WFS schema from the suite fixture.
     * 
     * @param testContext
     *            The test (group) context.
     */
    @BeforeClass
    public void setupClassFixture(ITestContext testContext) {
        this.wfsSchema = (Schema) testContext.getSuite().getAttribute(
                SuiteAttribute.WFS_SCHEMA.getName());
        Assert.assertNotNull(this.wfsSchema,
                "WFS schema not found in suite fixture.");
    }

    /**
     * Builds a DOM Document representing a GetFeature request entity that
     * contains no query expressions.
     */
    @BeforeMethod
    public void buildGetFeatureRequestEntity() {
        this.reqEntity = WFSRequest.createRequestEntity("GetFeature");
    }

    /**
     * [{@code Test}] If no stored query matches the given identifier then an
     * exception report with exception code "InvalidParameterValue" is expected.
     * 
     * @param binding
     *            The ProtocolBinding to use.
     */
    @Test(dataProvider = "protocol-binding")
    public void unknownQueryIdentifier(ProtocolBinding binding) {
        WFSRequest.appendStoredQuery(this.reqEntity,
                "http://docbook.org/ns/docbook",
                Collections.<String, Object> emptyMap());
        URI endpoint = ServiceMetadataUtils.getOperationEndpoint(
                this.wfsMetadata, WFS2.GET_FEATURE, binding);
        ClientResponse rsp = wfsClient.submitRequest(new DOMSource(
                this.reqEntity), binding, endpoint);
        Assert.assertTrue(rsp.hasEntity(),
                ErrorMessage.get(ErrorMessageKeys.MISSING_XML_ENTITY));
        Document entity = extractBodyAsDocument(rsp, binding);
        SchematronValidator validator = ValidationUtils
                .buildSchematronValidator("ExceptionReport.sch",
                        "InvalidParameterValuePhase");
        DOMResult result = validator.validate(new DOMSource(entity));
        Assert.assertFalse(validator.ruleViolationsDetected(), ErrorMessage
                .format(ErrorMessageKeys.NOT_SCHEMA_VALID,
                        validator.getRuleViolationCount(),
                        XMLUtils.writeNodeToString(result.getNode())));
    }

    /**
     * [{@code Test}] Invoking the {@code GetFeatureById} query with an
     * {@code id} parameter value that does not match any feature should produce
     * an error response with status code 404 (Not Found) or 403 (Forbidden).
     * 
     * Clause 11.4 stipulates that "In the event that a web feature service
     * encounters an error processing a GetFeature request, it shall raise an
     * OperationProcessingFailed exception as described in 7.5." In Table D.2
     * this exception code is mapped to status code 403 for some reason.
     * 
     * @param binding
     *            The ProtocolBinding to use.
     */
    @Test(dataProvider = "protocol-binding")
    public void invokeGetFeatureById_unknownID(ProtocolBinding binding) {
        String id = "uuid-" + UUID.randomUUID().toString();
        WFSRequest.appendStoredQuery(this.reqEntity,
                WFS2.QRY_GET_FEATURE_BY_ID,
                Collections.singletonMap("id", (Object) id));
        URI endpoint = ServiceMetadataUtils.getOperationEndpoint(
                this.wfsMetadata, WFS2.GET_FEATURE, binding);
        ClientResponse rsp = wfsClient.submitRequest(new DOMSource(
                this.reqEntity), binding, endpoint);
        int statusCode = rsp.getStatus();
        Assert.assertTrue(
                statusCode == ClientResponse.Status.NOT_FOUND.getStatusCode()
                        || statusCode == ClientResponse.Status.FORBIDDEN
                                .getStatusCode(),
                "Expected status code 403 or 404.");
    }

    /**
     * [{@code Test}] Invoking the {@code GetFeatureByType} query should produce
     * a result set containing feature members of the specified type. However,
     * instances need not exist for every supported feature type.
     * 
     * This test is primarily intended to acquire feature identifiers needed to
     * test the mandatory stored query <code>GetFeatureById</code>.
     * 
     * @param binding
     *            The ProtocolBinding to use.
     */
    @Test(dataProvider = "protocol-binding")
    public void invokeGetFeatureByType(ProtocolBinding binding) {
        NodeList members = null;
        for (QName featureType : this.featureTypes) {
            members = executeGetFeatureByType(binding, featureType);
            if ((null != members) && members.getLength() > 0) {
                break;
            }
        }
        Assert.assertNotNull(members,
                ErrorMessage.format(ErrorMessageKeys.NO_MEMBERS));
        for (int i = 0; i < members.getLength(); i++) {
            Element feature = (Element) members.item(i);
            String gmlId = feature.getAttributeNS(Namespaces.GML, "id");
            if (!gmlId.isEmpty()) {
                QName typeName = new QName(feature.getNamespaceURI(),
                        feature.getLocalName());
                this.featureIdentifiers.put(gmlId, typeName);
            }
        }
    }

    /**
     * [{@code Test}] Invoking the {@code GetFeatureById} query with a known
     * feature identifier should produce the matching feature representation
     * (@gml:id) as the response entity. If there is no matching feature, an
     * error response with a status code 404 (Not Found) is expected.
     * 
     * @param binding
     *            The ProtocolBinding to use.
     */
    @Test(dataProvider = "protocol-binding", dependsOnMethods = { "invokeGetFeatureByType" })
    public void invokeGetFeatureById(ProtocolBinding binding) {
        Map.Entry<String, QName> entry = this.featureIdentifiers.entrySet()
                .iterator().next();
        String featureId = entry.getKey();
        WFSRequest.appendStoredQuery(this.reqEntity,
                WFS2.QRY_GET_FEATURE_BY_ID,
                Collections.singletonMap("id", (Object) featureId));
        URI endpoint = ServiceMetadataUtils.getOperationEndpoint(
                this.wfsMetadata, WFS2.GET_FEATURE, binding);
        ClientResponse rsp = wfsClient.submitRequest(new DOMSource(
                this.reqEntity), binding, endpoint);
        Assert.assertTrue(rsp.hasEntity(),
                ErrorMessage.get(ErrorMessageKeys.MISSING_XML_ENTITY));
        Document entity = extractBodyAsDocument(rsp, binding);
        Element feature = entity.getDocumentElement();
        ETSAssert.assertQualifiedName(feature, entry.getValue());
        Assert.assertEquals(feature.getAttributeNS(Namespaces.GML, "id"),
                featureId, ErrorMessage.get(ErrorMessageKeys.UNEXPECTED_ID));
    }

    /**
     * Submits a request to execute the <code>GetFeatureByType</code> stored
     * query.
     * 
     * @param binding
     *            The ProtocolBinding (request encoding) to use.
     * @param featureType
     *            A QName specifying the feature type name.
     * @return A NodeList containing feature instances of the specified type (it
     *         may be empty).
     */
    NodeList executeGetFeatureByType(ProtocolBinding binding, QName featureType) {
        WFSRequest.appendStoredQuery(this.reqEntity,
                WFS2.QRY_GET_FEATURE_BY_TYPE,
                Collections.singletonMap("typeName", (Object) featureType));
        URI endpoint = ServiceMetadataUtils.getOperationEndpoint(
                this.wfsMetadata, WFS2.GET_FEATURE, binding);
        ClientResponse rsp = wfsClient.submitRequest(new DOMSource(
                this.reqEntity), binding, endpoint);
        Assert.assertEquals(rsp.getStatus(),
                ClientResponse.Status.OK.getStatusCode(),
                ErrorMessage.get(ErrorMessageKeys.UNEXPECTED_STATUS));
        Assert.assertTrue(rsp.hasEntity(),
                ErrorMessage.get(ErrorMessageKeys.MISSING_XML_ENTITY));
        Document entity = extractBodyAsDocument(rsp, binding);
        ETSAssert.assertQualifiedName(entity.getDocumentElement(), new QName(
                Namespaces.WFS, WFS2.FEATURE_COLLECTION));
        String xpath = "/wfs:FeatureCollection/wfs:member/*";
        Map<String, String> bindings = new HashMap<String, String>();
        bindings.put(Namespaces.WFS, "wfs");
        NodeList members;
        try {
            members = XMLUtils.evaluateXPath(entity, xpath, bindings);
        } catch (XPathExpressionException e) {
            throw new RuntimeException(e);
        }
        return members;
    }
}
